Merge branch 'dl/xdiff'
authorJunio C Hamano <junkio@cox.net>
Sat, 15 Apr 2006 05:58:17 +0000 (22:58 -0700)
committerJunio C Hamano <junkio@cox.net>
Sat, 15 Apr 2006 05:58:17 +0000 (22:58 -0700)
* dl/xdiff:
  xdiff: post-process hunks to make them consistent.

12 files changed:
Documentation/diff-options.txt
Documentation/git-log.txt
commit.c
diff.c
diff.h
git-diff.sh
git-fetch.sh
git-log.sh [deleted file]
git-parse-remote.sh
git-sh-setup.sh
stripspace.c
t/t5500-fetch-pack.sh

index 338014c..447e522 100644 (file)
@@ -7,6 +7,9 @@
 --patch-with-raw::
        Generate patch but keep also the default raw diff output.
 
+--stat::
+       Generate a diffstat instead of a patch.
+
 -z::
        \0 line termination on output
 
index e995d1b..76cb894 100644 (file)
@@ -12,11 +12,16 @@ SYNOPSIS
 
 DESCRIPTION
 -----------
-Shows the commit logs.  This command internally invokes
-'git-rev-list', and the command line options are passed to that
-command.
+Shows the commit logs.
+
+The command takes options applicable to the gitlink::git-rev-list[1]
+command to control what is shown and how, and options applicable to
+the gitlink::git-diff-tree[1] commands to control how the change
+each commit introduces are shown.
+
+This manual page describes only the most frequently used
+options.
 
-This manual page describes only the most frequently used options.
 
 OPTIONS
 -------
@@ -29,6 +34,12 @@ OPTIONS
 <since>..<until>::
        Show only commits between the named two commits.
 
+-p::
+       Show the change the commit introduces in a patch form.
+
+<paths>...::
+       Show only commits that affect the specified paths.
+
 
 Examples
 --------
@@ -47,6 +58,11 @@ git log --since="2 weeks ago" -- gitk::
        The "--" is necessary to avoid confusion with the *branch* named
        'gitk'
 
+git log -r --name-status release..test::
+
+       Show the commits that are in the "test" branch but not yet
+       in the "release" branch, along with the list of paths
+       each commit modifies.
 
 Author
 ------
index c7bb8db..ca25574 100644 (file)
--- a/commit.c
+++ b/commit.c
@@ -557,16 +557,11 @@ unsigned long pretty_print_commit(enum cmit_fmt fmt, const struct commit *commit
                if (fmt == CMIT_FMT_ONELINE)
                        break;
        }
-       if (fmt == CMIT_FMT_ONELINE) {
-               /* We do not want the terminating newline */
-               if (buf[offset - 1] == '\n')
-                       offset--;
-       }
-       else {
-               /* Make sure there is an EOLN */
-               if (buf[offset - 1] != '\n')
-                       buf[offset++] = '\n';
-       }
+       while (offset && isspace(buf[offset-1]))
+               offset--;
+       /* Make sure there is an EOLN for the non-oneline case */
+       if (fmt != CMIT_FMT_ONELINE)
+               buf[offset++] = '\n';
        buf[offset] = '\0';
        return offset;
 }
diff --git a/diff.c b/diff.c
index a14e664..f1b672d 100644 (file)
--- a/diff.c
+++ b/diff.c
@@ -8,7 +8,7 @@
 #include "quote.h"
 #include "diff.h"
 #include "diffcore.h"
-#include "xdiff/xdiff.h"
+#include "xdiff-interface.h"
 
 static int use_size_cache;
 
@@ -195,6 +195,137 @@ static int fn_out(void *priv, mmbuffer_t *mb, int nbuf)
        return 0;
 }
 
+struct diffstat_t {
+       struct xdiff_emit_state xm;
+
+       int nr;
+       int alloc;
+       struct diffstat_file {
+               char *name;
+               unsigned int added, deleted;
+       } **files;
+};
+
+static struct diffstat_file *diffstat_add(struct diffstat_t *diffstat,
+               const char *name)
+{
+       struct diffstat_file *x;
+       x = xcalloc(sizeof (*x), 1);
+       if (diffstat->nr == diffstat->alloc) {
+               diffstat->alloc = alloc_nr(diffstat->alloc);
+               diffstat->files = xrealloc(diffstat->files,
+                               diffstat->alloc * sizeof(x));
+       }
+       diffstat->files[diffstat->nr++] = x;
+       x->name = strdup(name);
+       return x;
+}
+
+static void diffstat_consume(void *priv, char *line, unsigned long len)
+{
+       struct diffstat_t *diffstat = priv;
+       struct diffstat_file *x = diffstat->files[diffstat->nr - 1];
+
+       if (line[0] == '+')
+               x->added++;
+       else if (line[0] == '-')
+               x->deleted++;
+}
+
+static const char pluses[] = "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++";
+static const char minuses[]= "----------------------------------------------------------------------";
+
+static void show_stats(struct diffstat_t* data)
+{
+       char *prefix = "";
+       int i, len, add, del, total, adds = 0, dels = 0;
+       int max, max_change = 0, max_len = 0;
+       int total_files = data->nr;
+
+       if (data->nr == 0)
+               return;
+
+       printf("---\n");
+
+       for (i = 0; i < data->nr; i++) {
+               struct diffstat_file *file = data->files[i];
+
+               if (max_change < file->added + file->deleted)
+                       max_change = file->added + file->deleted;
+               len = strlen(file->name);
+               if (max_len < len)
+                       max_len = len;
+       }
+
+       for (i = 0; i < data->nr; i++) {
+               char *name = data->files[i]->name;
+               int added = data->files[i]->added;
+               int deleted = data->files[i]->deleted;
+
+               if (0 < (len = quote_c_style(name, NULL, NULL, 0))) {
+                       char *qname = xmalloc(len + 1);
+                       quote_c_style(name, qname, NULL, 0);
+                       free(name);
+                       data->files[i]->name = name = qname;
+               }
+
+               /*
+                * "scale" the filename
+                */
+               len = strlen(name);
+               max = max_len;
+               if (max > 50)
+                       max = 50;
+               if (len > max) {
+                       char *slash;
+                       prefix = "...";
+                       max -= 3;
+                       name += len - max;
+                       slash = strchr(name, '/');
+                       if (slash)
+                               name = slash;
+               }
+               len = max;
+
+               /*
+                * scale the add/delete
+                */
+               max = max_change;
+               if (max + len > 70)
+                       max = 70 - len;
+
+               if (added < 0) {
+                       /* binary file */
+                       printf(" %s%-*s |  Bin\n", prefix, len, name);
+                       goto free_diffstat_file;
+               } else if (added + deleted == 0) {
+                       total_files--;
+                       goto free_diffstat_file;
+               }
+
+               add = added;
+               del = deleted;
+               total = add + del;
+               adds += add;
+               dels += del;
+
+               if (max_change > 0) {
+                       total = (total * max + max_change / 2) / max_change;
+                       add = (add * max + max_change / 2) / max_change;
+                       del = total - add;
+               }
+               printf(" %s%-*s |%5d %.*s%.*s\n", prefix,
+                               len, name, added + deleted,
+                               add, pluses, del, minuses);
+       free_diffstat_file:
+               free(data->files[i]->name);
+               free(data->files[i]);
+       }
+       free(data->files);
+       printf(" %d files changed, %d insertions(+), %d deletions(-)\n",
+                       total_files, adds, dels);
+}
+
 #define FIRST_FEW_BYTES 8000
 static int mmfile_is_binary(mmfile_t *mf)
 {
@@ -286,6 +417,35 @@ static void builtin_diff(const char *name_a,
        return;
 }
 
+static void builtin_diffstat(const char *name_a, const char *name_b,
+               struct diff_filespec *one, struct diff_filespec *two,
+               struct diffstat_t *diffstat)
+{
+       mmfile_t mf1, mf2;
+       struct diffstat_file *data;
+
+       data = diffstat_add(diffstat, name_a ? name_a : name_b);
+
+       if (fill_mmfile(&mf1, one) < 0 || fill_mmfile(&mf2, two) < 0)
+               die("unable to read files to diff");
+
+       if (mmfile_is_binary(&mf1) || mmfile_is_binary(&mf2))
+               data->added = -1;
+       else {
+               /* Crazy xdl interfaces.. */
+               xpparam_t xpp;
+               xdemitconf_t xecfg;
+               xdemitcb_t ecb;
+
+               xpp.flags = XDF_NEED_MINIMAL;
+               xecfg.ctxlen = 0;
+               xecfg.flags = 0;
+               ecb.outf = xdiff_outf;
+               ecb.priv = diffstat;
+               xdl_diff(&mf1, &mf2, &xpp, &xecfg, &ecb);
+       }
+}
+
 struct diff_filespec *alloc_filespec(const char *path)
 {
        int namelen = strlen(path);
@@ -819,6 +979,27 @@ static void run_diff(struct diff_filepair *p, struct diff_options *o)
        free(other_munged);
 }
 
+static void run_diffstat(struct diff_filepair *p, struct diff_options *o,
+               struct diffstat_t *diffstat)
+{
+       const char *name;
+       const char *other;
+
+       if (DIFF_PAIR_UNMERGED(p)) {
+               /* unmerged */
+               builtin_diffstat(p->one->path, NULL, NULL, NULL, diffstat);
+               return;
+       }
+
+       name = p->one->path;
+       other = (strcmp(name, p->two->path) ? p->two->path : NULL);
+
+       diff_fill_sha1_info(p->one);
+       diff_fill_sha1_info(p->two);
+
+       builtin_diffstat(name, other, p->one, p->two, diffstat);
+}
+
 void diff_setup(struct diff_options *options)
 {
        memset(options, 0, sizeof(*options));
@@ -866,6 +1047,8 @@ int diff_opt_parse(struct diff_options *options, const char **av, int ac)
                options->output_format = DIFF_FORMAT_PATCH;
                options->with_raw = 1;
        }
+       else if (!strcmp(arg, "--stat"))
+               options->output_format = DIFF_FORMAT_DIFFSTAT;
        else if (!strcmp(arg, "-z"))
                options->line_termination = 0;
        else if (!strncmp(arg, "-l", 2))
@@ -1160,11 +1343,24 @@ static void diff_flush_patch(struct diff_filepair *p, struct diff_options *o)
 
        if ((DIFF_FILE_VALID(p->one) && S_ISDIR(p->one->mode)) ||
            (DIFF_FILE_VALID(p->two) && S_ISDIR(p->two->mode)))
-               return; /* no tree diffs in patch format */ 
+               return; /* no tree diffs in patch format */
 
        run_diff(p, o);
 }
 
+static void diff_flush_stat(struct diff_filepair *p, struct diff_options *o,
+               struct diffstat_t *diffstat)
+{
+       if (diff_unmodified_pair(p))
+               return;
+
+       if ((DIFF_FILE_VALID(p->one) && S_ISDIR(p->one->mode)) ||
+           (DIFF_FILE_VALID(p->two) && S_ISDIR(p->two->mode)))
+               return; /* no tree diffs in patch format */
+
+       run_diffstat(p, o, diffstat);
+}
+
 int diff_queue_is_empty(void)
 {
        struct diff_queue_struct *q = &diff_queued_diff;
@@ -1276,7 +1472,8 @@ static void diff_resolve_rename_copy(void)
 
 static void flush_one_pair(struct diff_filepair *p,
                           int diff_output_format,
-                          struct diff_options *options)
+                          struct diff_options *options,
+                          struct diffstat_t *diffstat)
 {
        int inter_name_termination = '\t';
        int line_termination = options->line_termination;
@@ -1291,6 +1488,9 @@ static void flush_one_pair(struct diff_filepair *p,
                break;
        default:
                switch (diff_output_format) {
+               case DIFF_FORMAT_DIFFSTAT:
+                       diff_flush_stat(p, options, diffstat);
+                       break;
                case DIFF_FORMAT_PATCH:
                        diff_flush_patch(p, options);
                        break;
@@ -1316,19 +1516,31 @@ void diff_flush(struct diff_options *options)
        struct diff_queue_struct *q = &diff_queued_diff;
        int i;
        int diff_output_format = options->output_format;
+       struct diffstat_t *diffstat = NULL;
+
+       if (diff_output_format == DIFF_FORMAT_DIFFSTAT) {
+               diffstat = xcalloc(sizeof (struct diffstat_t), 1);
+               diffstat->xm.consume = diffstat_consume;
+       }
 
        if (options->with_raw) {
                for (i = 0; i < q->nr; i++) {
                        struct diff_filepair *p = q->queue[i];
-                       flush_one_pair(p, DIFF_FORMAT_RAW, options);
+                       flush_one_pair(p, DIFF_FORMAT_RAW, options, NULL);
                }
                putchar(options->line_termination);
        }
        for (i = 0; i < q->nr; i++) {
                struct diff_filepair *p = q->queue[i];
-               flush_one_pair(p, diff_output_format, options);
+               flush_one_pair(p, diff_output_format, options, diffstat);
                diff_free_filepair(p);
        }
+
+       if (diffstat) {
+               show_stats(diffstat);
+               free(diffstat);
+       }
+
        free(q->queue);
        q->queue = NULL;
        q->nr = q->alloc = 0;
diff --git a/diff.h b/diff.h
index 236095f..2f8aff2 100644 (file)
--- a/diff.h
+++ b/diff.h
@@ -119,6 +119,7 @@ extern void diffcore_std_no_resolve(struct diff_options *);
 "  -u            synonym for -p.\n" \
 "  --patch-with-raw\n" \
 "                output both a patch and the diff-raw format.\n" \
+"  --stat        show diffstat instead of patch.\n" \
 "  --name-only   show only names of changed files.\n" \
 "  --name-status show names and status of changed files.\n" \
 "  --full-index  show full object name on index lines.\n" \
@@ -142,6 +143,7 @@ extern int diff_queue_is_empty(void);
 #define DIFF_FORMAT_NO_OUTPUT  3
 #define DIFF_FORMAT_NAME       4
 #define DIFF_FORMAT_NAME_STATUS        5
+#define DIFF_FORMAT_DIFFSTAT   6
 
 extern void diff_flush(struct diff_options*);
 
index dc0dd31..0fe6770 100755 (executable)
@@ -30,9 +30,11 @@ case " $flags " in
        cc_or_p=--cc ;;
 esac
 
# If we do not have --name-status, --name-only, -r, or -c default to --cc.
 case " $flags " in
-*" '--name-status' "* | *" '--name-only' "* | *" '-r' "* | *" '-c' "* )
+*" '--name-status' "* | *" '--name-only' "* | *" '-r' "* | *" '-c' "* | \
+*" '--stat' "*)
        ;;
 *)
        flags="$flags'$cc_or_p' " ;;
index 711650f..83143f8 100755 (executable)
@@ -252,10 +252,10 @@ fetch_main () {
       else
          not_for_merge=
       fi
-      if expr "z$ref" : 'z\+' >/dev/null
+      if expr "z$ref" : 'z+' >/dev/null
       then
          single_force=t
-         ref=$(expr "z$ref" : 'z\+\(.*\)')
+         ref=$(expr "z$ref" : 'z+\(.*\)')
       else
          single_force=
       fi
@@ -30,9 +30,11 @@ case " $flags " in
 esac
 
-# If we do not have --name-status, --name-only, -r, or -c default to --cc.
+# If we do not have --name-status, --name-only, -r, -c or --stat,
+# default to --cc.
 case " $flags " in
diff --git a/git-log.sh b/git-log.sh
deleted file mode 100755 (executable)
index c2ea71c..0000000
+++ /dev/null
@@ -1,15 +0,0 @@
-#!/bin/sh
-#
-# Copyright (c) 2005 Linus Torvalds
-#
-
-USAGE='[--max-count=<n>] [<since>..<limit>] [--pretty=<format>] [git-rev-list options]'
-SUBDIRECTORY_OK='Yes'
-. git-sh-setup
-
-revs=$(git-rev-parse --revs-only --no-flags --default HEAD "$@") || exit
-[ "$revs" ] || {
-       die "No HEAD ref"
-}
-git-rev-list --pretty $(git-rev-parse --default HEAD "$@") |
-LESS=-S ${PAGER:-less}
index 65c66d5..c9b899e 100755 (executable)
@@ -77,7 +77,7 @@ canon_refs_list_for_fetch () {
                force=
                case "$ref" in
                +*)
-                       ref=$(expr "z$ref" : 'z\+\(.*\)')
+                       ref=$(expr "z$ref" : 'z+\(.*\)')
                        force=+
                        ;;
                esac
index 025ef2d..d15747f 100755 (executable)
@@ -30,7 +30,7 @@ $LONG_USAGE"
 fi
 
 case "$1" in
-       --h|--he|--hel|--help)
+       -h|--h|--he|--hel|--help)
        echo "$LONG_USAGE"
        exit
 esac
index dee1ef0..65a6346 100644 (file)
@@ -12,7 +12,9 @@ static int cleanup(char *line)
 {
        int len = strlen(line);
 
-       if (len > 1 && line[len-1] == '\n') {
+       if (len && line[len-1] == '\n') {
+               if (len == 1)
+                       return 0;
                do {
                        unsigned char c = line[len-2];
                        if (!isspace(c))
index e15e14f..92f12d9 100755 (executable)
@@ -12,11 +12,6 @@ test_description='Testing multi_ack pack fetching
 
 # Some convenience functions
 
-function show_count () {
-       commit_count=$(($commit_count+1))
-       printf "      %d\r" $commit_count
-}
-
 function add () {
        local name=$1
        local text="$@"
@@ -55,13 +50,6 @@ function test_expect_object_count () {
                "test $count = $output"
 }
 
-function test_repack () {
-       local rep=$1
-
-       test_expect_success "repack && prune-packed in $rep" \
-               '(git-repack && git-prune-packed)2>>log.txt'
-}
-
 function pull_to_client () {
        local number=$1
        local heads=$2
@@ -70,13 +58,23 @@ function pull_to_client () {
 
        cd client
        test_expect_success "$number pull" \
-               "git-fetch-pack -v .. $heads > log.txt 2>&1"
+               "git-fetch-pack -k -v .. $heads"
        case "$heads" in *A*) echo $ATIP > .git/refs/heads/A;; esac
        case "$heads" in *B*) echo $BTIP > .git/refs/heads/B;; esac
        git-symbolic-ref HEAD refs/heads/${heads:0:1}
+
        test_expect_success "fsck" 'git-fsck-objects --full > fsck.txt 2>&1'
-       test_expect_object_count "after $number pull" $count
-       pack_count=$(grep Unpacking log.txt|tr -dc "0-9")
+
+       test_expect_success 'check downloaded results' \
+       'mv .git/objects/pack/pack-* . &&
+        p=`ls -1 pack-*.pack` &&
+        git-unpack-objects <$p &&
+        git-fsck-objects --full'
+
+       test_expect_success "new object count after $number pull" \
+       'idx=`echo pack-*.idx` &&
+        pack_count=`git-show-index <$idx | wc -l` &&
+        test $pack_count = $count'
        test -z "$pack_count" && pack_count=0
        if [ -z "$no_strict_count_check" ]; then
                test_expect_success "minimal count" "test $count = $pack_count"
@@ -84,6 +82,7 @@ function pull_to_client () {
                test $count != $pack_count && \
                        echo "WARNING: $pack_count objects transmitted, only $count of which were needed"
        fi
+       rm -f pack-*
        cd ..
 }
 
@@ -117,8 +116,6 @@ git-symbolic-ref HEAD refs/heads/B
 
 pull_to_client 1st "B A" $((11*3))
 
-(cd client; test_repack client)
-
 add A11 $A10
 
 prev=1; cur=2; while [ $cur -le 65 ]; do
@@ -129,8 +126,6 @@ done
 
 pull_to_client 2nd "B" $((64*3))
 
-(cd client; test_repack client)
-
 pull_to_client 3rd "A" $((1*3)) # old fails
 
 test_done